import warnings
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional

from airphin.constants import CONFIG, REGEXP, TOKEN
from airphin.core.rules.loader import rule_calls, rule_imports
from airphin.utils.file import read_multi_yaml, read_yaml


class ParamDefaultConfig(NamedTuple):
    """Default statement config."""

    type: str
    value: str


class CallConfig(NamedTuple):
    """Call config."""

    long: str
    short: str
    src_long: str
    src_short: str
    replace: Dict[str, str]
    add: Dict[str, ParamDefaultConfig]
    remove: List[str]


class ImportStatementConfig(NamedTuple):
    """Import replace statement config."""

    inner_val: str
    inner_attr: str
    attr: str
    name: str
    statement: str


class ImportConfig(NamedTuple):
    """Import config."""

    replace: ImportStatementConfig
    add: List[str]
    remove: List[str]


class Config:
    """Configurations of airphin, including all configs change behavior of airphin.

    :param customs: User custom path of rules, will combine with build-in rules when :param:``customs_only``
        is False, will only use custom rules and ignore build-in rules when :param:``customs_only`` is True.
    :param customs_only: Only use custom rules or not.
    :param inplace: Replace source python file inplace instead of create a new file.
    :param imports: Build-in imports rules path.
    :param calls: Build-in call rules path.
    """

    def __init__(
        self,
        customs: Optional[List[Path]] = None,
        customs_only: Optional[bool] = False,
        inplace: Optional[bool] = False,
        imports: Optional[List[Path]] = rule_imports,
        calls: Optional[List[Path]] = rule_calls,
    ):
        self._customs = customs
        self.customs_only = customs_only
        if self.customs_only and not self._customs:
            raise ValueError(
                "Argument `customs` not allow value None, when customs_only is True."
            )
        if self.customs_only:
            warnings.warn(
                "Will only use customs rules to migration, will ignore built-in rules.",
                UserWarning,
                stacklevel=2,
            )
        self.inplace = inplace
        self._imports = imports
        self._calls = calls

    @property
    def imports_path(self) -> List[Path]:
        """Get all imports path for migration rules.

        Will only use :param:``customs`` rules and ignore built-in rules when :param:``customs_only`` is True,
        and combine :param:``customs``, built-in rules when :param:``customs_only`` is False.
        """
        if self.customs_only:
            return self._customs

        if not self._customs:
            return self._imports
        self._imports.extend(self._customs)
        return self._imports

    @property
    def imports(self) -> Dict[str, ImportConfig]:
        """Get all import migrator from rules."""
        return self.imp_migrator()

    @property
    def calls_path(self) -> List[Path]:
        """Get all call path for migration rules.

        Will only use :param:``customs`` rules and ignore built-in rules when :param:``customs_only`` is True,
        and combine :param:``customs``, built-in rules when :param:``customs_only`` is False.
        """
        if self.customs_only:
            return self._customs

        if not self._customs:
            return self._calls
        self._calls.extend(self._customs)
        return self._calls

    @property
    def calls(self) -> Dict[str, CallConfig]:
        """Get all call migrator from rules."""
        return self.call_migrator()

    @staticmethod
    def _build_caller(
        src: str, dest: str, parameters: List[Dict[str, Any]]
    ) -> CallConfig:
        return CallConfig(
            long=dest,
            short=dest.split(TOKEN.POINT)[-1],
            src_long=src,
            src_short=src.split(TOKEN.POINT)[-1],
            replace={
                p[CONFIG.SOURCE]: p[CONFIG.DESTINATION]
                for p in parameters
                if p[CONFIG.ACTION] == CONFIG.KW_REPLACE
            },
            add={
                p[CONFIG.ARGUMENT]: ParamDefaultConfig(
                    type=p[CONFIG.DEFAULT][CONFIG.TYPE],
                    value=p[CONFIG.DEFAULT][CONFIG.VALUE],
                )
                for p in parameters
                if p[CONFIG.ACTION] == CONFIG.KW_ADD
            },
            remove=[
                p[CONFIG.ARGUMENT]
                for p in parameters
                if p[CONFIG.ACTION] == CONFIG.KW_REMOVE
            ],
        )

    @staticmethod
    def _handle_rules_path(*args) -> List[Dict[str, Any]]:
        for arg in args:
            if arg.is_file():
                content: Dict = read_yaml(arg)
                yield content
            if arg.is_dir():
                contents: List[Dict] = read_multi_yaml(arg.rglob(REGEXP.PATH_YAML))
                for content in contents:
                    yield content

    @staticmethod
    def get_module_action(
        migration: Dict[str, Any], action_type: str
    ) -> Optional[Dict[str, Any]]:
        """Get specific action type from rules.

        :param migration: Config Migration node.
        :param action_type: Action type, can be `add`, `remove`, `replace`.
        """
        actions = [
            action
            for action in migration[CONFIG.MODULE]
            if action[CONFIG.ACTION] == action_type
        ]
        if len(actions) > 1:
            raise ValueError("Each type of action can only have one.")
        return actions[0] if actions else None

    def call_migrator(self) -> Dict[str, CallConfig]:
        """Get all call migrator from rules."""
        migrator = {}

        for content in self._handle_rules_path(*self.calls_path):
            migration = content[CONFIG.MIGRATION]
            parameters = migration[CONFIG.PARAMETER]
            replace = self.get_module_action(migration, CONFIG.KW_REPLACE)
            src = replace[CONFIG.SOURCE]
            dest = replace[CONFIG.DESTINATION]

            if isinstance(src, str):
                migrator[src] = self._build_caller(src, dest, parameters)
            elif isinstance(src, list):
                for inner_src in src:
                    migrator[inner_src] = self._build_caller(
                        inner_src, dest, parameters
                    )
            else:
                raise RuntimeError("Invalid migration.module.src type: %s" % type(src))
        return migrator

    # TODO: make it can use in any number module import statemant
    @staticmethod
    def _build_replace_importer(action: Dict[str, Any]) -> ImportStatementConfig:
        dest = action[CONFIG.DESTINATION]
        inner_val, inner_attr, attr, name = dest.split(TOKEN.POINT)
        return ImportStatementConfig(
            inner_val=inner_val,
            inner_attr=inner_attr,
            attr=attr,
            name=name,
            statement=f"from {inner_val}.{inner_attr}.{attr} import {name}",
        )

    @staticmethod
    def _get_rp_add_action(action: Dict[str, Any]) -> Optional[List[str]]:
        """Get replace and add action list from rules.

        :param action: Config migration module action.
        """

        def _build_import_statement(mod: str) -> str:
            spec = mod.rsplit(TOKEN.POINT, 1)
            return f"from {spec[0]} import {spec[1]}"

        if action is None:
            return []
        module = action[CONFIG.MODULE]
        if isinstance(module, str):
            return [_build_import_statement(module)]
        elif isinstance(module, list):
            return [_build_import_statement(mod) for mod in module]
        else:
            raise RuntimeError(
                "Invalid migration.module.action.module type: %s" % type(module)
            )

    def imp_migrator(self) -> Dict[str, ImportConfig]:
        """Get all import migrator from rules."""
        imps = {}

        for content in self._handle_rules_path(*self.imports_path):
            replace = self.get_module_action(
                content[CONFIG.MIGRATION], CONFIG.KW_REPLACE
            )
            add = self.get_module_action(content[CONFIG.MIGRATION], CONFIG.KW_ADD)
            remove = self.get_module_action(content[CONFIG.MIGRATION], CONFIG.KW_REMOVE)

            src = replace[CONFIG.SOURCE]
            if isinstance(src, str):
                imps[src] = ImportConfig(
                    replace=self._build_replace_importer(replace),
                    add=self._get_rp_add_action(add),
                    remove=self._get_rp_add_action(remove),
                )
            elif isinstance(src, list):
                for inner_src in src:
                    imps[inner_src] = ImportConfig(
                        replace=self._build_replace_importer(replace),
                        add=self._get_rp_add_action(add),
                        remove=self._get_rp_add_action(remove),
                    )

        return imps



class TOKEN:
    """Constants token for airphin."""

    IMPORT: str = "import"
    COMMA: str = ","
    POINT: str = "."
    STRING: str = "str"
    CODE: str = "code"
    NEW_LINE: str = "\n"


class KEYWORD:
    """Constants keywords for airphin."""

    MIGRATE_MARK: str = "-airphin"


class REGEXP:
    """Constants regular expression for airphin."""

    PATH_YAML: str = "*.yaml"
    PATH_PYTHON: str = "*.py"
    ALL: str = "**/*"


class CONFIG:
    """Constants config file for airphin."""

    EXAMPLE: str = "examples"

    MIGRATION: str = "migration"
    MODULE: str = "module"
    PARAMETER: str = "parameter"

    ACTION: str = "action"
    SOURCE: str = "src"
    DESTINATION: str = "dest"

    KW_REPLACE: str = "replace"
    KW_ADD: str = "add"
    KW_REMOVE: str = "remove"

    ARGUMENT: str = "arg"
    DEFAULT: str = "default"
    TYPE: str = "type"
    VALUE: str = "value"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Optional, Sequence, Union

import libcst as cst
import libcst.matchers as m
from libcst import FlattenSentinel, RemovalSentinel

from airphin.constants import TOKEN
from airphin.core.rules.config import Config, ImportConfig


class ImportTransformer(cst.CSTTransformer):
    """CST Transformer for airflow operators."""

    def __init__(self, config: Config):
        super().__init__()
        self.config: Config = config
        self.mod_ref = None
        self.class_names = []

    def _get_attr_nested_value(self, node: cst.Attribute) -> str:
        if m.matches(node.value, m.TypeOf(m.Name)):
            return f"{node.value.value}.{node.attr.value}"
        elif m.matches(node.value, m.TypeOf(m.Attribute)):
            nested = self._get_attr_nested_value(
                cst.ensure_type(node.value, cst.Attribute)
            )
            return f"{nested}.{node.attr.value}"

    def visit_ImportFrom(self, node: cst.ImportFrom) -> Optional[bool]:
        # case ``from modules import class``
        if m.matches(node.module, m.TypeOf(m.Name)):
            self.mod_ref = node.module.value
        # case ``from package.module.[module1] import class``
        elif m.matches(node.module, m.TypeOf(m.Attribute)):
            self.mod_ref = self._get_attr_nested_value(
                cst.ensure_type(node.module, cst.Attribute)
            )

        # skip ``import *``, aka ImportStar
        if isinstance(node.names, Sequence):
            self.class_names = [
                cst.ensure_type(ia, cst.ImportAlias).name.value for ia in node.names
            ]
        return False

    def leave_ImportFrom(
        self, original_node: cst.ImportFrom, updated_node: cst.ImportFrom
    ) -> Union[
        cst.BaseSmallStatement, FlattenSentinel[cst.BaseSmallStatement], RemovalSentinel
    ]:
        if self.mod_ref is not None:
            src_full_refs = [
                f"{self.mod_ref}.{class_name}" for class_name in self.class_names
            ]

            replaces = []
            adds = set()
            for full_ref in src_full_refs:
                if full_ref in self.config.imports:
                    dest: ImportConfig = self.config.imports[full_ref]
                    replaces.append(dest.replace.statement)
                    adds.update(dest.add)

            # get replace statement
            if len(replaces) == 0:
                return updated_node
            elif len(replaces) == 1:
                statement = replaces[0]
            else:
                class_name_only = [
                    stat.split(f" {TOKEN.IMPORT} ")[1] for stat in replaces[1:]
                ]
                statement = f"{TOKEN.COMMA} ".join(replaces[:1] + class_name_only)

            # Return replace and add statement
            # TODO, will use ; as separator of multiple statements, we should better use \n in the future
            return FlattenSentinel(
                [
                    *[cst.parse_statement(add).body[0] for add in adds],
                    cst.parse_statement(statement).body[0],
                ]
            )
        return updated_node



from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml

from airphin.constants import REGEXP


def read(path: Path) -> str:
    """Read content from path.

    :param path: Path to read content.
    """
    with open(path, "r") as f:
        return f.read()


def write(path: Path, content: str) -> None:
    """Write content to path.

    The path's content will be overwritten if they are already exists.

    :param path: Path to write content.
    :param content: Content want to write to path.
    """
    with open(path, "w") as f:
        f.write(content)


def read_yaml(path: Path) -> Dict[str, Any]:
    """Read yaml file and return a dict.

    :param path: Path to read content.
    """
    assert path.is_file(), "Path must be a single file."
    content = read(path)
    return yaml.safe_load(content)


def read_multi_yaml(paths: List[Path]) -> List[Dict[str, Any]]:
    """Read multiple yaml files and return a list of dict.

    :param paths: List of paths to read.
    """
    yamls = []
    for path in paths:
        content = read_yaml(path)
        yamls.append(content)
    return yamls


def add_stem_suffix(path: Path, suf: str) -> Path:
    """Add stem suffix of path.

    This function add suffix to stem instead of suffix of path, for example:

    >>> add_stem_suffix(Path("foo/bar/baz.py"), "_test")
    Path("foo/bar/baz_test.py")

    :param path: Path to add suffix.
    :param suf: Suffix want to add to stem.
    """
    stem, suffix = path.stem, path.suffix
    new_name = f"{stem}{suf}{suffix}"
    return path.with_name(new_name)


def recurse_files(
    path: Path, include: Optional[str] = REGEXP.ALL, exclude: Optional[str] = None
) -> List[Path]:
    """Recurse all match pattern files in path.

    :param path: file or directory path want to recurse.
    :param include: include match pattern in given path, default include all file in directory.
    :param exclude: include match pattern in given path, default None.
    """
    if not path.exists():
        raise ValueError("Path %s does not exist.", path)

    if path.is_file():
        return [path]
    else:
        paths = set(path.rglob(include))
        if exclude:
            paths = paths - set(path.rglob(exclude))
        return [file for file in paths if file.is_file()]



from pathlib import Path
from typing import List

project = Path(__file__).parent.parent.parent

path_rule = project.joinpath("rules")
path_operators = path_rule.joinpath("operators")
path_dag_cnx = path_rule.joinpath("core", "dagContext.yaml")

rule_imports = [path_operators, path_dag_cnx]

rule_calls = [
    path_operators,
    path_dag_cnx,
]


def build_in_rules() -> List[Path]:
    """Get all build-in rules in airphin.rules directory."""
    return [path for path in path_rule.glob("**/*") if path.is_file()]



import logging
import multiprocessing
from multiprocessing import Pool
from pathlib import Path
from timeit import default_timer as timer
from typing import List, Optional

import libcst as cst
from tqdm import tqdm

from airphin.constants import KEYWORD, TOKEN
from airphin.core.rules.config import Config
from airphin.core.transformer.route import Transformer
from airphin.utils.file import add_stem_suffix, read, write

logger = logging.getLogger("airphin.runner")


class Runner:
    """Airphin runner, main class to run transformer.

    :param config: Config of airphin.
    """

    def __init__(self, config: Config) -> None:
        self.config: Config = config

    def with_str(self, content: str) -> str:
        """Run airphin with a string and return migrated content.

        :param content: Content of string you want to migrate.
        """
        parse_cst = cst.parse_module(content)
        wrapper = cst.MetadataWrapper(parse_cst)
        migrated = wrapper.visit(Transformer(self.config)).code
        return migrated

    def with_file(self, path: Path) -> None:
        """Run airphin with a single file path and migrate to dolphinscheduler python sdk definition.

        Will change file inplace when ``config.inplace = True``, and create new file end with ``-airphin``
        when ``config.inplace = False``.

        :param path: Path of file you want to migrate.
        """
        logger.debug("Start migrate file %s", path)
        start = timer()
        content = read(path)
        migrated = self.with_str(content)

        if self.config.inplace:
            write(path, migrated)
        else:
            new_path = add_stem_suffix(path, KEYWORD.MIGRATE_MARK)
            write(new_path, migrated)

        logger.debug("End migrate file %s, elapsed time %.5fs", path, timer() - start)

    def with_files(self, paths: List[Path]) -> None:
        """Run airphin with multiple files to dolphinscheduler python sdk definition.

        :param paths: Path of file you want to migrate.
        """
        logger.info("Start migrate files, total %d files scan.", len(paths))
        logger.debug(
            "Start migrate files, files contain:\n%s",
            TOKEN.NEW_LINE.join((f"  {p}" for p in paths)),
        )

        start = timer()
        for file in tqdm(paths):
            self.with_file(file)

        logger.info(
            f"Total migrated {len(paths)} files, spend time: %.5fs.", timer() - start
        )

    def with_files_multiprocess(
        self, paths: List[Path], processes: Optional[int] = multiprocessing.cpu_count()
    ) -> None:
        """Run airphin migrating with multiprocess.

        :param paths: Path of file you want to migrate.
        :param processes: multiprocess processes cpu count number.
        """
        logger.info(
            "Start multiple processing migrate files, total %d files scan.", len(paths)
        )
        logger.debug(
            "Start migrate files with processes number %d, files contain:\n%s",
            processes,
            TOKEN.NEW_LINE.join((f"  {p}" for p in paths)),
        )

        start = timer()
        with Pool(processes) as pool:
            list(tqdm(pool.imap(self.with_file, paths), total=len(paths)))

        logger.debug(
            "All files had add to multiprocess pool, spend time %.5fs.", timer() - start
        )
        pool.join()

        logger.info(
            f"Total migrated {len(paths)} files, spend time: %.5fs.", timer() - start
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import warnings
from typing import Set, Union

import libcst as cst
from libcst import BaseExpression, FlattenSentinel, RemovalSentinel
from libcst.metadata import PositionProvider, QualifiedName, QualifiedNameProvider

from airphin.core.rules.config import Config
from airphin.core.transformer.imports import ImportTransformer
from airphin.core.transformer.operators import OpTransformer


class Transformer(cst.CSTTransformer):
    """CST Transformer route class from airflow to dolphinscheduler-sdk-python.

    The main class to call each rules to migrate, just like a router, currently will route to `imports` and
    `operators` transformer.
    """

    METADATA_DEPENDENCIES = (
        QualifiedNameProvider,
        PositionProvider,
    )

    def __init__(self, config: Config):
        super().__init__()
        self.config: Config = config

    @staticmethod
    def _get_qualified_name(qualifie: Set[QualifiedName]) -> str:
        if len(qualifie) > 1:
            warnings.warn(
                "QualifiedNameProvider get more than one qualified name, will use the first one.",
                RuntimeWarning,
            )
        for q in qualifie:
            return q.name

    # @m.call_if_inside(
    #     m.Call(
    #         func=m.MatchMetadataIfTrue(
    #             meta.QualifiedNameProvider,
    #             lambda qualnames: any(
    #                 n.name in call_cov for n in qualnames
    #             ),
    #         )
    #     )
    # )
    def leave_Call(
        self, original_node: cst.Call, updated_node: cst.Call
    ) -> BaseExpression:
        qnp = self.get_metadata(QualifiedNameProvider, original_node)
        if qnp:
            qnp_name = self._get_qualified_name(qnp)
            if qnp_name in self.config.calls:
                return updated_node.visit(OpTransformer(self.config, qnp_name))
        return updated_node

    def leave_ImportFrom(
        self, original_node: cst.ImportFrom, updated_node: cst.ImportFrom
    ) -> Union[
        cst.BaseSmallStatement, FlattenSentinel[cst.BaseSmallStatement], RemovalSentinel
    ]:
        """Migrate from import statement."""
        return updated_node.visit(ImportTransformer(self.config))



import argparse
import difflib
import logging
import sys
from pathlib import Path
from typing import Dict, Sequence

from airphin import __project_name__, __version__
from airphin.constants import REGEXP, TOKEN
from airphin.core.rules.config import Config
from airphin.core.rules.loader import build_in_rules, path_rule
from airphin.runner import Runner
from airphin.utils.file import recurse_files

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger("airphin")

common_args: Dict[str, Dict] = {
    "custom_rules": {
        "help": f"The custom rule file path you want to add to {__project_name__}.",
        "action": "append",
        "type": Path,
    },
    "custom_only": {
        "help": "Only use custom rules and ignore all built-in's, it is helpful for patching the"
        "exists migration.",
        "action": "store_true",
    },
    "verbose": {
        "action": "store_true",
        "help": "Show more verbose output.",
    },
}


def build_argparse() -> argparse.ArgumentParser:
    """Build argparse.ArgumentParser with specific configuration."""
    parser = argparse.ArgumentParser(
        prog="airphin",
        description="Airphin is a tool for migrating Airflow DAGs to DolphinScheduler Python API.",
    )

    # Version
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"{__project_name__} version {__version__}",
        help="Show version of %(prog)s.",
    )

    # Subcommands
    subparsers = parser.add_subparsers(
        title="subcommands",
        dest="subcommand",
        help=f"Subcommand you want to {__project_name__} to run.",
    )

    # Test
    parser_test = subparsers.add_parser(
        "test", help=f"{__project_name__} playground for migrating with standard input."
    )
    parser_test.add_argument(
        "-v",
        "--verbose",
        **common_args["verbose"],
    )
    parser_test.add_argument(
        "-r",
        "--custom-rules",
        **common_args["custom_rules"],
    )
    parser_test.add_argument(
        "-R",
        "--custom-only",
        **common_args["custom_only"],
    )
    parser_test.add_argument(
        "-d",
        "--diff",
        action="store_true",
        help=f"Prints diff of all the changes {__project_name__} would make.",
    )
    parser_test.add_argument(
        "stdin",
        help="The standard input you want to migrate.",
        action="store",
        type=str,
    )

    # migrate
    parser_migrate = subparsers.add_parser(
        "migrate", help="Migrate Airflow DAGs to DolphinScheduler Python definition."
    )
    parser_migrate.add_argument(
        "-v",
        "--verbose",
        **common_args["verbose"],
    )
    parser_migrate.add_argument(
        "-r",
        "--custom-rules",
        **common_args["custom_rules"],
    )
    parser_migrate.add_argument(
        "-R",
        "--custom-only",
        **common_args["custom_only"],
    )
    parser_migrate.add_argument(
        "-I",
        "--include",
        help=f"Include files based on conditions provided, default '{REGEXP.PATH_PYTHON}'",
        action="store",
        default=REGEXP.PATH_PYTHON,
        type=str,
    )
    parser_migrate.add_argument(
        "-E",
        "--exclude",
        help="Exclude files based on conditions provided, without default value",
        action="store",
        type=str,
    )
    parser_migrate.add_argument(
        "-i",
        "--inplace",
        help="Migrate python file in place instead of create a new file.",
        action="store_true",
    )
    parser_migrate.add_argument(
        "-m",
        "--multiprocess",
        help="Migrate python files with multiprocess.",
        action="store",
        type=int,
    )
    parser_migrate.add_argument(
        "sources",
        default=[Path(".")],
        nargs="*",
        help="The directories or files paths you want to migrate.",
        action="store",
        type=Path,
    )

    # Rule
    parser_rule = subparsers.add_parser("rule", help="Rule of migrating.")
    parser_rule.add_argument(
        "-s",
        "--show",
        action="store_true",
        help=f"Show all rules for {__project_name__} migrate.",
    )

    return parser


def main(argv: Sequence[str] = None) -> None:
    """Run airphin in command line."""
    parser = build_argparse()
    argv = argv if argv is not None else sys.argv[1:]
    args = parser.parse_args(argv)

    if hasattr(args, "verbose") and args.verbose:
        logger.setLevel(logging.DEBUG)
    logger.debug("Finish parse airphin arguments, current args is %s.", args)

    # recurse all file in given path
    customs_rules = []
    if hasattr(args, "custom_rules") and args.custom_rules:
        for rule in args.custom_rules:
            customs_rules.extend(recurse_files(rule))
    if logger.level <= logging.DEBUG and customs_rules:
        logger.debug(
            "This migration have custom rules:\n%s",
            TOKEN.NEW_LINE.join((f"  {r}" for r in customs_rules)),
        )

    if args.subcommand == "test":
        stdin = args.stdin
        config = Config(customs=customs_rules, customs_only=args.custom_only)
        runner = Runner(config)

        result = runner.with_str(stdin)
        logger.debug("The source input is:\n%s", stdin)
        logger.info(f"Migrated result is: \n{result}")

        if args.diff:
            diff = difflib.unified_diff(
                stdin.splitlines(keepends=True),
                result.splitlines(keepends=True),
                fromfile="source",
                tofile="dest",
            )
            logger.info(
                f"The different between source and target is: \n{''.join(diff)}"
            )

    if args.subcommand == "migrate":
        migrate_files = []
        for path in args.sources:
            migrate_files.extend(recurse_files(path, args.include, args.exclude))

        config = Config(
            customs=customs_rules, customs_only=args.custom_only, inplace=args.inplace
        )
        runner = Runner(config)

        if args.multiprocess:
            runner.with_files_multiprocess(migrate_files, args.multiprocess)
        else:
            runner.with_files(migrate_files)

    if args.subcommand == "rule":
        if args.show:
            rules = build_in_rules()
            logger.info(f"Total {len(rules)} rules:\n")
            for rule in rules:
                print(rule.relative_to(path_rule))


if __name__ == "__main__":
    raise SystemExit(main())



import copy
from typing import Optional, Sequence, Union

import libcst as cst
import libcst.matchers as m
from libcst import Arg, BaseExpression, FlattenSentinel, RemovalSentinel

from airphin.constants import TOKEN
from airphin.core.rules.config import CallConfig, Config, ParamDefaultConfig


class OpTransformer(cst.CSTTransformer):
    """CST Transformer for airflow operators.

    TODO Need to skip inner call like DAG(date_time=datetime.datetime.now().strftime("%Y-%m-%d"))

    :param qualified_name: qualified name of operator
    """

    def __init__(self, config: Config, qualified_name: Optional[str] = None):
        super().__init__()
        self._config: Config = config
        self.qualified_name = qualified_name
        assert self.qualified_name is not None
        self.visit_name = False
        self.migrated_param = set()

    @property
    def config(self) -> CallConfig:
        return self._config.calls.get(self.qualified_name)

    def matcher_op_name(self, node: cst.Name) -> bool:
        if self.visit_name is False and node.value == self.config.src_short:
            self.visit_name = True
            return True
        return False

    def match_replace_name(self, node: cst.Arg) -> bool:
        migrate_names = self.config.replace.keys()
        return m.matches(
            node,
            m.Arg(keyword=m.Name(m.MatchIfTrue(lambda name: name in migrate_names))),
        )

    def match_remove_name(self, node: cst.Arg) -> bool:
        return m.matches(
            node,
            m.Arg(
                keyword=m.Name(m.MatchIfTrue(lambda name: name in self.config.remove))
            ),
        )

    def match_call_name(self, node: cst.Call) -> bool:
        if m.matches(node.func, m.TypeOf(m.Name)):
            val = cst.ensure_type(node.func, cst.Name).value
        elif m.matches(node.func, m.TypeOf(m.Attribute)):
            val = cst.ensure_type(node.func, cst.Attribute).attr.value
        else:
            return True
        return val in self.qualified_name.split(TOKEN.POINT)

    def leave_Name(
        self, original_node: cst.Name, updated_node: cst.Name
    ) -> "BaseExpression":
        if self.matcher_op_name(original_node):
            dest_name = self.config.short
            return updated_node.with_changes(value=dest_name)
        return updated_node

    def leave_Arg(
        self, original_node: cst.Arg, updated_node: cst.Arg
    ) -> Union[Arg, FlattenSentinel[cst.Arg], RemovalSentinel]:
        if self.match_replace_name(original_node):
            original_keyword = original_node.keyword.value
            dest_keyword = self.config.replace.get(original_keyword)

            self.migrated_param.add(dest_keyword)
            return updated_node.with_changes(keyword=cst.Name(value=dest_keyword))
        if self.match_remove_name(original_node):
            return cst.RemoveFromParent()
        return updated_node

    def _handle_missing_default(self, nodes: Sequence[cst.Arg]) -> Sequence[cst.Arg]:
        mutable = list(nodes)
        one_of = copy.deepcopy(mutable[-1])
        for arg in self.config.add.keys():
            default: ParamDefaultConfig = self.config.add.get(arg)

            if default.type == TOKEN.STRING:
                value = cst.SimpleString(value=f'"{default.value}"')
            elif default.type == TOKEN.CODE:
                value = cst.parse_expression(default.value)
            else:
                raise NotImplementedError
            mutable.append(
                one_of.with_changes(
                    value=value,
                    keyword=cst.Name(value=arg),
                )
            )
        return mutable

    def leave_Call(
        self, original_node: cst.Call, updated_node: cst.Call
    ) -> BaseExpression:
        if not self.config.add:
            return updated_node

        if not self.match_call_name(original_node):
            return updated_node

        return updated_node.with_changes(
            args=self._handle_missing_default(updated_node.args)
        )
